
/*
   CV2MIDI 4U Synth control voltage to MIDI conterter

  FOR TEENSY 4.0 and ADS8681

   2 module channels (called devices here) translate input voltage to MIDI out.
   For each analog input there is an input attenuator and offset pot.
   Switches select MIDI channel and MIDI continuous control for each device.

   Bothe Devices can transmit on MIDI channels 1-8, individually selected

   EXCEPT:
   With Device B set to GATE and Device A set to NOTES:
   - Device B will transmit on Device A channel
   - Device B will output a MIDI NOTE OFF when it's input goes below aprox 2.5v, and a MIDI note on
     if the gate in goes above aprox 2.5v and gate was previously off and it's not a new note.

  Pitch Bend is a 14 bit signal.

  The following also transmit as 14 bit. Whether or not it's recieved as 14 bit depends on the destination.
  1. Volume on both devices
  2. CC 12/ MIDI EFFECT 1
  3. CC 13/ MIDI EFFECT 2

  All other selections are 7 Bit (0 to 127)

  When changing the switches for MIDI channel or MIDI continuous control
  the previous controller will be set to it's default value if the RTZ (return to zero) switch in on.
  If the switch is off the controller will be left at it's last setting. This can cause notes to hang.

   Device A CC switch selections
   0 - Notes
   1 - Pitch Bend (2 byte)
   2 - Modulation
   3 - Volume (2 byte)
   4 - Pan
   5 - Continuous Control 12 (MIDI effect 1) (2 byte)
   6 - Continuous Control 14
   7 - Continuous Control 15

   Device B CC switch selections
   0 - GATE: Sends Note On/Off messages for Device A notes
   1 - Pitch Bend (2 byte)
   2 - Volume (2 byte)
   3 - Pan
   4 - Continuous Control 13 (MIDI effect 2) (2 byte)
   5 - Continuous Control 16
   6 - Continuous Control 17
   7 - Continuous Control 18

*/

// 2 channel CV to MIDI converter
// Craig Lee 2025

#include <SPI.h>

// pin assignments
const int MidOut = 1; // MIDI TX (Serial1.write)
const int ACh_0 = 2; // A Channel bit 0 ****** Device A ******
const int ACh_2 = 3; // A Channel bit 2
const int ACh_4 = 4; // A Channel bit 4
const int ACC_0 = 5; // A Continuous Control bit 0
const int ACC_2 = 6; // A Continuous Control bit 2
const int ACC_4 = 7; // A Continuous Control bit 4
const int InSelect = 9; // select ADC input 1 or 2
const int ADC_CS = 10; // external DAC chip select ****** ADC ******
const int ADC_MOSI = 11; // ADC input
const int ADC_MISO = 12; // ADC output
const int ADC_SCK = 13; // ADC clock
const int RTZ = 14; // return to zero switch
const int BCh_0 = 16; // B Channel bit 0 ****** Device B ******
const int BCh_2 = 17; // B Channel bit 2
const int BCh_4 = 18; // B Channel bit 4
const int BCC_0 = 19; // B Continuous Control bit 0
const int BCC_2 = 20; // B Continuous Control bit 2
const int BCC_4 = 21; // B Continuous Control bit 4
const int ALED = 22; // activity LED CH A pin
const int BLED = 23; // activity LED CH B pin
//-------------------------------------------------------------------
// variables
bool device_IN = 0; // set which INPUT is read by ADC and updated
byte channel_A; // A what MIDI channel to transmit on
byte channel_B; // B what MIDI channel to transmit on
byte contControl_A; // A what continuous controller to transmit
byte contControl_B; // B what continuous controller to transmit
byte old_channel_A; // storage for old settings
byte old_channel_B;
byte old_cc_A;
byte old_cc_B;
unsigned int newVoltage; // result of new ADC conversion
unsigned int old_A_conversion; // stored A conversion
unsigned int old_B_conversion; // stored B conversion
byte A_data; // A data to send
byte B_data; // B data to send
byte old_A_data; // A data last sent
byte old_B_data; // B data last sent
byte A_highData; // A high byte for Pitch Bend and 14 bits
byte A_lowData; // A low byte for Pitch Bend and 14 bits
byte B_highData; // B high byte for Pitch Bend and 14 bits
byte B_lowData; // B low byte for Pitch Bend and 14 bits
byte send_A; // A device built channel and controller message
byte send_B; // B device built channel and controller message
unsigned int noteDebounce; // "debounce" new note
bool gateStatus; // for gate to note off, is gate on or off
bool RTZ_SW = 1; // return to zero when 1
byte A_LED; // flag that A LED is on
byte B_LED; // flag that B LED is on
unsigned int A_delay; // count down to turn LED off
unsigned int B_delay; // count down to turn LED off
unsigned int LEDdelay = 3500; // how many microseconds to light LEDs
int change; // used for hysterisis routines - reused, don't save
unsigned int flag = 50; // change needed to do math - hysterisis amount
// switch debouncing
byte readchA; // read to see if switch is changed
byte readchB; // read to see if switch is changed
byte readccA; // read to see if switch is changed
byte readccB; // read to see if switch is changed
unsigned int CCA_switch; // used for debouncing switches
unsigned int CCB_switch; // used for debouncing switches
unsigned int CHA_switch; // used for debouncing switches
unsigned int CHB_switch; // used for debouncing switches
bool debounce_ChA = 0; // debounce flag
bool debounce_ChB = 0; // debounce flag
bool debounce_ccA = 0; // debounce flag
bool debounce_ccB = 0; // debounce flag

byte waitStartup = 0; // run program loop a few times before outputting data
long scaleRead; // make up for in not full scale
unsigned int A; // throwaway variable for printout page

//-------------------------------------------------------------------


// SETUP
void setup()
{
  setupPins();

  SPI.begin();
  digitalWriteFast(ADC_CS, HIGH);
  delay(10);
  digitalWriteFast(ALED, HIGH);
  digitalWriteFast(BLED, HIGH);

  // set adc input range
  SPI.beginTransaction(SPISettings(10000000, MSBFIRST, SPI_MODE0)); //
  digitalWriteFast(ADC_CS, LOW);
  delayMicroseconds(50);
  SPI.transfer16(0b1101000000010100); // write to RANGE_SEL_REG
  SPI.transfer16(0b0000000000001011); // set range to 1.25XVref (5.11875)
  SPI.endTransaction();
  digitalWriteFast(ADC_CS, HIGH);

  Serial1.begin(31250); // setup serial for MIDI on TX PIN

  // inital read of switches
  channel_A = digitalReadFast(ACh_0) | (digitalReadFast(ACh_2) << 1) | (digitalReadFast(ACh_4) << 2);
  channel_B = digitalReadFast(BCh_0) | (digitalReadFast(BCh_2) << 1) | (digitalReadFast(BCh_4) << 2);
  contControl_A = digitalReadFast(ACC_0) | (digitalReadFast(ACC_2) << 1) | (digitalReadFast(ACC_4) << 2);
  contControl_B  = digitalReadFast(BCC_0) | (digitalReadFast(BCC_2) << 1) | (digitalReadFast(BCC_4) << 2);

  delay (500);
  digitalWriteFast(ALED, LOW);
  digitalWriteFast(BLED, LOW);

  device_IN = 1;
  digitalWriteFast(InSelect, device_IN); // set multiplexer


}




//-------------LOOP-------------------------

void loop()
{
  device_IN = !device_IN; // flip
  //  digitalWriteFast(InSelect, device_IN); // set multiplexer moved
  delayMicroseconds(50); // wait switching time
  digitalWriteFast(ADC_CS, HIGH); // start conversion
  delayMicroseconds(2); // wait conversion

  // Get switch settings - read switches
  // if any have changed, zero out CCs and turn off notes
  // so old controllers are at zero and notes don't hang
  // before switching

  old_channel_A = channel_A;
  old_channel_B = channel_B;
  old_cc_A = contControl_A;
  old_cc_B = contControl_B;
  RTZ_SW = digitalReadFast(RTZ); // read return to zero switch

  // Read the switches and debounce:
  // If switch changed, wait the debounce time before reading.
  // Switches are break before make so you need to wait that out.

  //----- UNIT A MIDI CHANNEL ------
  if (debounce_ChA == 0) // read the switch to check if it's changed
  {
    readchA = digitalReadFast(ACh_0) | (digitalReadFast(ACh_2) << 1) | (digitalReadFast(ACh_4) << 2);
  }

  if (readchA != old_channel_A & debounce_ChA == 0) // If change, start debounce countdown
  {
    debounce_ChA = 1;
    CHA_switch = millis();
  }

  if (debounce_ChA == 1)
  {
    if (millis() - CHA_switch > 300) // after countdown read switch
    {
      debounce_ChA = 0;
      channel_A = digitalReadFast(ACh_0) | (digitalReadFast(ACh_2) << 1) | (digitalReadFast(ACh_4) << 2);
      readchA = channel_A;
    }
  }
  //----- UNIT B MIDI CHANNEL ------
  if (debounce_ChB == 0)
  {
    readchA = digitalReadFast(BCh_0) | (digitalReadFast(BCh_2) << 1) | (digitalReadFast(BCh_4) << 2);
  }

  if (readchA != old_channel_B & debounce_ChB == 0)
  {
    debounce_ChB = 1;
    CHB_switch = millis();
  }

  if (debounce_ChB == 1)
  {
    if (millis() - CHB_switch > 300)
    {
      debounce_ChB = 0;
      channel_B = digitalReadFast(BCh_0) | (digitalReadFast(BCh_2) << 1) | (digitalReadFast(BCh_4) << 2);
      readchB = channel_B;
    }
  }
  //----- UNIT A Continuous Controll ------
  if (debounce_ccA == 0)
  {
    readccA = digitalReadFast(ACC_0) | (digitalReadFast(ACC_2) << 1) | (digitalReadFast(ACC_4) << 2);
  }

  if (readccA != old_cc_A & debounce_ccA == 0)
  {
    debounce_ccA = 1;
    CCA_switch = millis();
  }

  if (debounce_ccA == 1)
  {
    if (millis() - CCA_switch > 300)
    {
      debounce_ccA = 0;
      contControl_A = digitalReadFast(ACC_0) | (digitalReadFast(ACC_2) << 1) | (digitalReadFast(ACC_4) << 2);
      readccA = contControl_A;
    }
  }
  //----- UNIT B Continuous Controll ------
  if (debounce_ccB == 0)
  {
    readccB = digitalReadFast(BCC_0) | (digitalReadFast(BCC_2) << 1) | (digitalReadFast(BCC_4) << 2);
  }

  if (readccB != old_cc_B & debounce_ccB == 0)
  {
    debounce_ccB = 1;
    CCB_switch = millis();
  }

  if (debounce_ccB == 1)
  {
    if (millis() - CCB_switch > 300)
    {
      debounce_ccB = 0;
      contControl_B  = digitalReadFast(BCC_0) | (digitalReadFast(BCC_2) << 1) | (digitalReadFast(BCC_4) << 2);
      readccB = contControl_B;
    }
  }

  // Have any switches changed?
  // if so zero out old settings if RTZ is on

  if (RTZ_SW == 1)
  {
    if (channel_A != old_channel_A)
    {
      A_device_zero();
    }

    if (old_cc_A != contControl_A)
    {
      A_device_zero();
    }

    if (channel_B != old_channel_B)
    {
      B_device_zero();
    }

    if (old_cc_B != contControl_B)
    {
      B_device_zero();
    }
  }

  // if Device B is set to gate (switch 0) the unit B MIDI channel matches unit A
  if (contControl_B == 0 & contControl_A == 0)
  {
    channel_B = channel_A;
  }
  // read a CV and send

  // ADC read
  SPI.beginTransaction(SPISettings(30000000, MSBFIRST, SPI_MODE0));
  digitalWriteFast(ADC_CS, LOW);
  newVoltage = SPI.transfer16(0);
  SPI.endTransaction();

  // switch multiplexer for next sample
  digitalWriteFast(InSelect, !device_IN); // set multiplexer

  //  multiplier insures full scale read
  //  ADC referance voltage is 5.12v so 5.0v wouldn't output 127
  scaleRead = newVoltage * 1.03;
  if (scaleRead > 65535)
  {
    scaleRead = 65535;
  }
  newVoltage = scaleRead;

  // Switch device_IN selects which device is being update
  switch (device_IN)
  {

    default: // 0: A device
      change = old_A_conversion - newVoltage; // hysterisis

      switch (contControl_A) // what controller is sent for A
      {
        default: //notes
          if (abs(change) > 500) // added noise immunity
          {
            old_A_conversion = newVoltage;
          }
          else
          {
            newVoltage = old_A_conversion;
          }
          A_oneByte();

          break;

        case 1: //Pitech Bend
          if (abs(change) > flag) // hysterisis for pitch bend
          {
            old_A_conversion = newVoltage;
          }
          else
          {
            newVoltage = old_A_conversion;
          }
          change = (newVoltage >> 2) - 8192; // get dead band at zero
          if (abs(change) < 160)
          {
            newVoltage = 32768;
          }
          A_highData = newVoltage >> 9; // create 2 byte data
          A_lowData = (newVoltage >> 2) & 127;
          newVoltage = A_lowData;
          break;

        case 2: //Modulation
          A_oneByte(); // figures data to send and stores conversion as old
          break;

        case 3: //Volume
          A_twoByte();
          break;

        case 4: //Pan
          A_oneByte();
          break;

        case 5: //CC12 (effect 1)
          A_twoByte();
          break;

        case 6: //CC14
          A_oneByte();
          break;

        case 7: //CC15 14 bit
          A_oneByte();
          break;
      }

      // send data
      if (newVoltage != old_A_data)
      {
        if (contControl_A == 0) // output notes
        {
          //gateStatus = 1; // a new note turns on gate
          noteDebounce = millis();
        }
        digitalWriteFast(ALED, HIGH);
        A_LED = HIGH;
        A_delay = micros();
        A_data = newVoltage;
        if (waitStartup > 12) //loop a few times before outputting data
        {
          A_channel_send(); // go output data
        }
        else
        {
          waitStartup++;
        }

        old_A_data = A_data;
      }

      break; // case 0/ default ---------------------


    case 1: // B device
      change = old_B_conversion - newVoltage; // hysterisis

      switch (contControl_B) // what controller is sent for B
      {
        default: // Gates
          if (abs(change) > 800) // added noise immunity
          {
            old_B_conversion = newVoltage;
          }
          else
          {
            newVoltage = old_B_conversion;
          }
          B_oneByte();
          newVoltage = newVoltage >> 6; // make one bit
          break;

        case 1: //Pitech Bend
          if (abs(change) > flag) // hysterisis for pitch bend
          {
            old_B_conversion = newVoltage;
          }
          else
          {
            newVoltage = old_B_conversion;
          }
          change = (newVoltage >> 2) - 8192; // get dead band at zero
          if (abs(change) < 160)
          {
            newVoltage = 32768;
          }
          B_highData = newVoltage >> 9; // create 2 byte data
          B_lowData = (newVoltage >> 2) & 127;
          newVoltage = B_lowData;
          break;

        case 2: //Volume
          B_twoByte();
          break;

        case 3: //Pan
          B_oneByte();
          break;

        case 4: //CC13 (effect 2)
          B_twoByte();
          break;

        case 5: //CC16
          B_oneByte();
          break;

        case 6: //CC17 14 bit
          B_oneByte();
          break;

        case 7: //CC18 14 bit
          B_oneByte();
          break;
      }

      //Send Data
      if (newVoltage != old_B_data)
      {
        digitalWriteFast(BLED, HIGH);
        B_LED = HIGH;
        B_delay = micros();
        B_data = newVoltage;
        if (waitStartup > 12) //loop a few times before outputting data
        {
          B_channel_send(); // go output data
        }
        else
        {
          waitStartup++;
        }
        B_channel_send(); // go output data
        old_B_data = B_data;
      }
      break; // case 1 -----------------------

  } // end - switch device_IN ----------------------


  // LED handlers
  if (A_LED = HIGH)
  {
    if (micros() - A_delay > LEDdelay)
    {
      digitalWriteFast(ALED, LOW);
      A_LED = LOW;
    }
  }

  if (B_LED = HIGH)
  {
    if (micros() - B_delay > LEDdelay)
    {
      digitalWriteFast(BLED, LOW);
      B_LED = LOW;
    }
  }

  //    printout(); // printout info to serial monitor

}

// END OF MAIN LOOP------------------------------------------------

void A_oneByte()
{
  if (abs(change) > flag) // 256 is half a step/note
  {
    old_A_conversion = newVoltage;
  }
  else
  {
    newVoltage = old_A_conversion;
  }
  newVoltage = newVoltage >> 9; // make sample 7 bit
}

void B_oneByte()
{
  if (abs(change) > flag) // 256 is half a step/note
  {
    old_B_conversion = newVoltage;
  }
  else
  {
    newVoltage = old_B_conversion;
  }
  newVoltage = newVoltage >> 9; // make sample 7 bit
}

void A_twoByte()
{
  if (abs(change) > flag) // hysterisis for pitch bend ******too high******
  {
    old_A_conversion = newVoltage;
  }
  else
  {
    newVoltage = old_A_conversion;
  }
  A_highData = newVoltage >> 9;
  A_lowData = (newVoltage >> 2) & 127;

  if (A_highData == 0 & A_lowData < flag)
  {
    A_lowData = 0;
  }
  newVoltage = A_lowData;
}

void B_twoByte()
{
  if (abs(change) > flag) // hysterisis for pitch bend ******too high******
  {
    old_B_conversion = newVoltage;
  }
  else
  {
    newVoltage = old_B_conversion;
  }
  B_highData = newVoltage >> 9;
  B_lowData = (newVoltage >> 2) & 127;

  if (B_highData == 0 & B_lowData < flag)
  {
    B_lowData = 0;
  }
  newVoltage = B_lowData;
}

//------------------END--------------------
